import os
from logging import Logger
from threading import Lock

from datacube.utils.aws import configure_s3_access

credential_lock = Lock()


class CredentialManager:
    _instance = None

    def __new__(cls, log: Logger | None = None) -> "CredentialManager":
        # new/init assumed to be called with credential_lock held
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, log: Logger | None = None) -> None:
        # new/init assumed to be called with credential_lock held
        # Startup initialisation of libraries controlled by environment variables
        self.use_aws = False
        self.unsigned = False
        self.requester_pays = False
        self.credentials = None
        self.log = log

        if log:
            log.debug("Initialising CredentialManager")

        # Boto3/AWS
        if os.environ.get("AWS_DEFAULT_REGION") or os.environ.get("AWS_REGION"):
            if "AWS_NO_SIGN_REQUEST" in os.environ:
                env_nosign = os.environ["AWS_NO_SIGN_REQUEST"]
                if env_nosign.lower() in ("y", "t", "yes", "true", "1"):
                    unsigned = True
                    # Workaround for rasterio bug
                    os.environ["AWS_NO_SIGN_REQUEST"] = "yes"
                    os.environ["AWS_ACCESS_KEY_ID"] = "fake"
                    os.environ["AWS_SECRET_ACCESS_KEY"] = "fake"
                else:
                    unsigned = False
                    # delete env variable
                    del os.environ["AWS_NO_SIGN_REQUEST"]
            else:
                unsigned = False
                if log:
                    log.warning(
                        "AWS_NO_SIGN_REQUEST is not set. "
                        + "The default behaviour has recently changed to False (i.e. signed requests) "
                        + "Please explicitly set $AWS_NO_SIGN_REQUEST to 'no' for unsigned requests."
                    )
            env_requester_pays = os.environ.get("AWS_REQUEST_PAYER", "")
            requester_pays = False
            if env_requester_pays.lower() == "requester":
                requester_pays = True
            self.use_aws = True
            if log:
                if unsigned:
                    log.info("S3 access configured with unsigned requests")
                else:
                    log.info("S3 access configured with signed requests")
            self.unsigned = unsigned
            self.requester_pays = requester_pays
            self.renew_creds()

            if "AWS_S3_ENDPOINT" in os.environ and os.environ["AWS_S3_ENDPOINT"] == "":
                del os.environ["AWS_S3_ENDPOINT"]
        elif log:
            log.warning(
                "Environment variable $AWS_DEFAULT_REGION not set.  (This warning can be ignored if all data is stored locally.)"
            )

    def _check_cred(self) -> None:
        from botocore.credentials import RefreshableCredentials

        if self.credentials and isinstance(self.credentials, RefreshableCredentials):
            if self.credentials.refresh_needed():
                self.renew_creds()
            elif self.log:
                # pylint: disable=protected-access
                self.log.info(
                    "Credentials look OK: %s seconds remaining",
                    str(self.credentials._seconds_remaining()),
                )
        elif self.log:
            self.log.debug(
                "Credentials of type %s - NOT RENEWING",
                self.credentials.__class__.__name__,
            )

    @classmethod
    def check_cred(cls) -> None:
        # pylint: disable=protected-access
        with credential_lock:
            assert cls._instance is not None
            cls._instance._check_cred()

    def renew_creds(self) -> None:
        if self.use_aws:
            from botocore.credentials import RefreshableCredentials

            if self.log:
                self.log.info("Establishing/renewing credentials")
            self.credentials = configure_s3_access(
                aws_unsigned=self.unsigned, requester_pays=self.requester_pays
            )
            if self.log and isinstance(self.credentials, RefreshableCredentials):
                # pylint: disable=protected-access
                self.log.debug(
                    "%s seconds remaining", str(self.credentials._seconds_remaining())
                )


def initialise_aws_credentials(log: Logger | None = None) -> None:
    # pylint: disable=protected-access
    with credential_lock:
        if CredentialManager._instance is None:
            _ = CredentialManager(log)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import warnings
from collections.abc import Callable
from logging import Logger
from time import monotonic
from typing import Any, Optional, TypeVar

from flask import Flask, g, request

TYPE_CHECKING = False
if TYPE_CHECKING:
    from datacube_ows.ows_configuration import OWSConfig

__all__ = [
    "create_app",
]


def initialise_logger(name: str | None = None) -> Logger:
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("[%(asctime)s] [%(levelname)s] %(message)s"))
    log = logging.getLogger(name)
    log.addHandler(handler)
    # If invoked using Gunicorn, link our root logger to the gunicorn logger
    # this will mean the root logs will be captured and managed by the gunicorn logger
    # allowing you to set the gunicorn log directories and levels for logs
    # produced by this application
    log.setLevel(logging.getLogger("gunicorn.error").getEffectiveLevel())
    return log


def initialise_ignorable_warnings() -> None:
    # Suppress annoying rasterio warning message every time we write to a non-georeferenced image format
    from rasterio.errors import NotGeoreferencedWarning

    warnings.simplefilter("ignore", category=NotGeoreferencedWarning)


def initialise_debugging(log: Logger | None = None) -> None:
    # PYCHARM Debugging
    dbg = os.environ.get("PYDEV_DEBUG")
    if dbg and dbg.lower() not in ("no", "false", "f", "n"):
        import pydevd_pycharm

        pydevd_pycharm.settrace(
            "172.17.0.1", port=12321, stdout_to_server=True, stderr_to_server=True
        )
        if log:
            log.info("PyCharm Debugging enabled")


SentryEvent = TypeVar("SentryEvent")


def before_send(event: SentryEvent, hint: dict[str, Any]) -> SentryEvent | None:
    if "exc_info" in hint:
        _, exc_value, _ = hint["exc_info"]
        if isinstance(
            exc_value, AttributeError
        ) and "object has no attribute 'GEOSGeom_destroy'" in str(exc_value):
            return None
    return event


def initialise_sentry(log: Logger | None = None) -> None:
    if os.environ.get("SENTRY_DSN"):
        import sentry_sdk
        from sentry_sdk.integrations.flask import FlaskIntegration

        SENTRY_ENV_TAG = os.environ.get("SENTRY_ENV_TAG") or "dev"

        sentry_sdk.init(
            dsn=os.environ["SENTRY_DSN"],
            environment=SENTRY_ENV_TAG,
            integrations=[FlaskIntegration()],
            before_send=before_send,
        )
        if log:
            log.info("Sentry initialised")


def parse_config_file(log: Logger | None = None) -> Optional["OWSConfig"]:
    # Cache a parsed config file object
    # (unless deferring to first request)
    from datacube_ows.ows_configuration import get_config

    cfg = None
    if not os.environ.get("DEFER_CFG_PARSE"):
        cfg = get_config()
    return cfg


def initialise_flask(name: str) -> Flask:
    app_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    return Flask(
        name.split(".")[0], template_folder=os.path.join(app_path, "templates")
    )


def pass_through(undecorated: Callable) -> Callable:
    def decorator(*args, **kwargs):
        return undecorated(*args, **kwargs)

    decorator.__name__ = undecorated.__name__
    return decorator


class FakeMetrics:
    def do_not_track(self) -> Callable:
        return pass_through

    def counter(self, *args, **kwargs) -> Callable:
        return pass_through

    def histogram(self, *args, **kwargs) -> Callable:
        return pass_through

    def gauge(self, *args, **kwargs) -> Callable:
        return pass_through

    def summary(self, *args, **kwargs) -> Callable:
        return pass_through


def initialise_prometheus(app: Flask, log: Logger | None = None):
    # Prometheus
    if os.environ.get("PROMETHEUS_MULTIPROC_DIR", False):
        from prometheus_flask_exporter.multiprocess import (
            GunicornInternalPrometheusMetrics,
        )

        metrics = GunicornInternalPrometheusMetrics(app, group_by="endpoint")
        if log:
            log.info("Prometheus metrics enabled")
        return metrics
    return FakeMetrics()


def proxy_fix(app: Flask, log: Logger | None = None):
    # Proxy Fix, to respect X-Forwarded-For headers
    if os.environ.get("PROXY_FIX", False):
        from werkzeug.middleware.proxy_fix import ProxyFix

        app.wsgi_app = ProxyFix(app.wsgi_app, x_for=1, x_proto=1, x_host=1, x_port=1)  # type: ignore[method-assign]
        if log is not None:
            log.info("ProxyFix was enabled")
    return app


def request_extractor() -> str | None:
    return request.args.get("request")


def initialise_babel(cfg, app: Flask) -> object | None:
    if cfg and cfg.internationalised:
        from flask_babel import Babel

        app.config["BABEL_TRANSLATION_DIRECTORIES"] = cfg.translations_dir

        def get_locale() -> str | None:
            return request.accept_languages.best_match(
                cfg.locales, default=cfg.locales[0]
            )

        return Babel(
            app,
            locale_selector=get_locale,
            default_domain=cfg.message_domain,
            configure_jinja=False,
        )
    return None


def create_app() -> Flask:
    log = initialise_logger()
    app = initialise_flask(__name__)
    cfg = parse_config_file(log)
    initialise_ignorable_warnings()
    initialise_debugging(log)
    initialise_sentry(log)
    from datacube_ows.startup_utils.creds import initialise_aws_credentials

    initialise_aws_credentials(log)
    initialise_babel(cfg, app)
    metrics = initialise_prometheus(app, log)
    ows_ogc_metric = metrics.histogram(
        "ows_ogc",
        "Summary by OGC request protocol, version, operation, layer, and HTTP Status",
        labels={
            "query_request": lambda: request.args.get("request", "NONE").upper(),
            "query_service": lambda: request.args.get("service", "NONE").upper(),
            "query_version": lambda: request.args.get("version"),
            "query_layer": lambda: (
                request.args.get("query_layers")  # WMS GetFeatureInfo
                or request.args.get("layers")  # WMS
                or request.args.get("layer")  # WMTS
                or request.args.get("coverage")  # WCS 1.x
                or request.args.get("coverageid")  # WCS 2.x
            ),
            "status": lambda r: r.status_code,
        },
    )
    proxy_fix(app, log)

    # Declare routes
    from datacube_ows.ogc import (
        legend,
        ogc_impl,
        ogc_wcs_impl,
        ogc_wms_impl,
        ogc_wmts_impl,
        ping,
    )

    @app.route("/")
    @ows_ogc_metric
    def main_route():
        return ogc_impl()

    @app.route("/wms")
    @ows_ogc_metric
    def wms_route():
        return ogc_wms_impl()

    @app.route("/wmts")
    @ows_ogc_metric
    def wmts_route():
        return ogc_wmts_impl()

    @app.route("/wcs")
    @ows_ogc_metric
    def wcs_route():
        return ogc_wcs_impl()

    @app.route("/ping")
    @metrics.summary(
        "ows_heartbeat_pings", "Ping durations", labels={"status": lambda r: r.status}
    )
    def ping_route():
        return ping()

    @app.route("/legend/<string:layer>/<string:style>/legend.png")
    @metrics.histogram(
        "ows_legends",
        "Legend query durations",
        labels={
            "layer": lambda: request.path.split("/")[2],
            "style": lambda: request.path.split("/")[3],
            "status": lambda r: r.status,
        },
    )
    def legend_route(layer, style):
        return legend(layer, style)

    # Configure Flask Middleware
    @app.before_request
    def start_timer() -> None:
        # pylint: disable=assigning-non-slot
        g.ogc_start_time = monotonic()

    @app.after_request
    def log_time_and_request_response(response):
        time_taken = int((monotonic() - g.ogc_start_time) * 1000)
        # request.environ.get('HTTP_X_REAL_IP') captures requester ip on a local docker container via gunicorn
        if request.environ.get("HTTP_X_REAL_IP"):
            ip = request.environ.get("HTTP_X_REAL_IP")
        # request.environ.get('HTTP_X_FORWARDED_FOR') captures request IP forwarded by ingress/loadbalancer
        elif request.environ.get("HTTP_X_FORWARDED_FOR"):
            ip = request.environ.get("HTTP_X_FORWARDED_FOR")
        # request.environ.get('REMOTE_ADDR') is standard internal IP address
        elif request.environ.get("REMOTE_ADDR"):
            ip = request.environ.get("REMOTE_ADDR")
        else:
            ip = "Not found"
        log.info(
            "ip: %s request: %s returned status: %d and took: %d ms",
            ip,
            request.url,
            response.status_code,
            time_taken,
        )
        return response

    return app



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

from collections.abc import Iterable
from threading import Lock
from typing import Literal, cast
from uuid import UUID

import click
from datacube import Datacube
from datacube.drivers.common_psql import as_role
from datacube.model import Dataset, Product
from odc.geo import CRS, Geometry
from sqlalchemy import text
from typing_extensions import override

from datacube_ows.index.api import (
    InsufficientDbPrivileges,
    LayerExtent,
    LayerSignature,
    OWSAbstractIndex,
    OWSAbstractIndexDriver,
    TimeSearchTerm,
    check_perms,
)
from datacube_ows.index.sql import run_sql
from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.utils import get_driver_name, get_sqlconn

from .mv_index import MVSelectOpts, mv_search
from .product_ranges import create_range_entry as create_range_entry_impl
from .product_ranges import get_ranges as get_ranges_impl


class OWSPostgresIndex(OWSAbstractIndex):
    name: str = "postgres"

    # method to check database access (for ping op)
    @override
    def check_db_access(self, dc: Datacube) -> bool:
        db_ok = False
        try:
            with get_sqlconn(dc) as conn:
                results = conn.execute(
                    text("""
                    SELECT *
                    FROM ows.layer_ranges
                    LIMIT 1""")
                )
                for _ in results:
                    db_ok = True
        except Exception:
            pass
        return db_ok

    @override
    def _check_perms(self, dc: Datacube, group: Literal["manage", "admin"]) -> None:
        if get_driver_name(dc.index) == "psycopg":
            from psycopg.errors import ProgrammingError
        else:
            from psycopg2.errors import ProgrammingError
        try:
            with get_sqlconn(dc) as conn, as_role(conn, f"agdc_{group}"):
                pass
        except ProgrammingError as e:
            raise InsufficientDbPrivileges(
                f"db user {dc.index.environment.db_username} does not have agdc_{group} privileges: {e}"
            ) from None

    # method to delete obsolete schemas etc.
    @override
    @check_perms("admin")
    def cleanup_schema(self, dc: Datacube) -> None:
        self._run_sql(dc, "ows_schema/cleanup")

    # Schema creation method
    @override
    @check_perms("admin")
    def create_schema(self, dc: Datacube) -> None:
        click.echo("Creating schema and postgis extension...")
        self._run_sql(dc, "ows_schema/bootstrap")
        click.echo("Creating/updating tables...")
        self._run_sql(dc, "ows_schema/create")
        click.echo("Creating/updating materialised views...")
        self._run_sql(dc, "extent_views/create")
        click.echo("Granting tables permissions to agdc roles.")
        self._run_sql(dc, "ows_schema/grants")
        click.echo("Granting views permissions to agdc roles.")
        self._run_sql(dc, "extent_views/grants")

    # Spatiotemporal index update method (e.g. refresh materialised views)
    @override
    @check_perms("manage")
    def update_geotemporal_index(self, dc: Datacube) -> None:
        self._run_sql(dc, "extent_views/refresh")

    @override
    @check_perms("manage")
    def create_range_entry(
        self, layer: OWSNamedLayer, cache: dict[LayerSignature, list[str]]
    ) -> None:
        create_range_entry_impl(layer, cache)

    @override
    def get_ranges(self, layer: OWSNamedLayer) -> LayerExtent | None:
        return get_ranges_impl(layer)

    @override
    def ds_search(
        self,
        layer: OWSNamedLayer,
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
    ) -> Iterable[Dataset]:
        return cast(
            Iterable[Dataset],
            mv_search(
                layer.dc,
                MVSelectOpts.DATASETS,
                times=times,
                geom=geom,
                products=products,
            ),
        )

    @override
    def dsid_search(
        self,
        layer: OWSNamedLayer,
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
    ) -> Iterable[UUID]:
        return cast(
            Iterable[UUID],
            mv_search(
                layer.dc,
                MVSelectOpts.IDS,
                times=times,
                geom=geom,
                products=products,
            ),
        )

    @override
    def count(
        self,
        layer: OWSNamedLayer,
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
    ) -> int:
        return cast(
            int,
            mv_search(
                layer.dc,
                MVSelectOpts.COUNT,
                times=times,
                geom=geom,
                products=products,
            ),
        )

    @override
    def extent(
        self,
        layer: OWSNamedLayer,
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
        crs: CRS | None = None,
    ) -> Geometry | None:
        extent = cast(
            Geometry | None,
            mv_search(
                layer.dc,
                MVSelectOpts.EXTENT,
                times=times,
                geom=geom,
                products=products,
            ),
        )
        if extent is None or crs is None or crs == extent.crs:
            return extent
        return extent.to_crs(crs)

    def _run_sql(self, dc: Datacube, path: str, **params: str) -> bool:
        return run_sql(dc, path, **params)


pgdriverlock = Lock()


class OWSPostgresIndexDriver(OWSAbstractIndexDriver):
    _driver = None

    @classmethod
    @override
    def ows_index_class(cls) -> type[OWSAbstractIndex]:
        return OWSPostgresIndex

    @classmethod
    @override
    def ows_index(cls) -> OWSAbstractIndex:
        with pgdriverlock:
            if cls._driver is None:
                cls._driver = OWSPostgresIndex()
        return cls._driver


def ows_index_driver_init() -> OWSPostgresIndexDriver:
    return OWSPostgresIndexDriver()



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging
import math
from collections.abc import Callable
from datetime import UTC, date, datetime

import click
import datacube
import odc.geo
from odc.geo.crs import CRS
from sqlalchemy import text

from datacube_ows.index.api import CoordRange, LayerExtent, LayerSignature
from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.utils import get_driver_name, get_sqlconn

_LOG: logging.Logger = logging.getLogger(__name__)


def jsonise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    return {
        "top": bbox.top,
        "bottom": bbox.bottom,
        "left": bbox.left,
        "right": bbox.right,
    }


def create_range_entry(
    layer: OWSNamedLayer, cache: dict[LayerSignature, list[str]]
) -> None:
    # Take layer signature
    meta = LayerSignature(
        time_res=layer.time_resolution.value,
        products=tuple(layer.product_names),
        env=layer.local_env._name,
        datasets=layer.dc.index.datasets.count(product=layer.product_names),
    )

    with get_sqlconn(layer.dc) as conn:
        txn = conn.begin()
        cached = cache.get(meta)
        if get_driver_name(layer.dc.index) == "psycopg":
            from psycopg.types.json import Json
        else:
            from psycopg2.extras import Json
        # insert initial empty row if one does not already exist, so we can safely update it later.
        conn.execute(
            text("""
                 INSERT INTO ows.layer_ranges
                 (layer, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated)
                 VALUES (:p_layer, 0, 0, 0, 0, :empty, :empty, :meta, :now)
                 ON CONFLICT (layer) DO NOTHING
                 """),
            {
                "p_layer": layer.name,
                "empty": Json(""),
                "meta": Json(meta.as_json()),
                "now": datetime.now(tz=UTC),
            },
        )
        if cached is not None:
            template = cached[0]
            # Layer {template} has same signature - reusing
            conn.execute(
                text("""
            UPDATE ows.layer_ranges lr1
            SET lat_min = lr2.lat_min,
                lat_max = lr2.lat_max,
                lon_min = lr2.lon_min,
                lon_max = lr2.lon_max,
                dates = lr2.dates,
                bboxes = lr2.bboxes,
                meta = lr2.meta,
                last_updated = lr2.last_updated
            FROM ows.layer_ranges lr2
            WHERE lr1.layer = :layer_id
            AND   lr2.layer = :template_id"""),
                {"layer_id": layer.name, "template_id": template},
            )
        else:
            prodids = [p.id for p in layer.products]

            # Set default timezone
            conn.execute(text("""set timezone to 'Etc/UTC'"""))

            # Loop over dates
            dates = set()  # Should get to here!
            if layer.time_resolution.is_solar():
                results = conn.execute(
                    text(
                        """
                  select lower(dt.search_val), upper(dt.search_val),
                         lower(lon.search_val), upper(lon.search_val)
                  from odc.dataset ds, odc.dataset_search_datetime dt, odc.dataset_search_num lon
                  where ds.product_ref = ANY(:prodids)
                  AND ds.id = dt.dataset_ref
                  AND ds.id = lon.dataset_ref
                  AND dt.search_key = :time
                  AND lon.search_key = :lon
                  """
                    ),
                    {"prodids": prodids, "time": "time", "lon": "lon"},
                )
                for result in results:
                    dt1, dt2, ll, lu = result
                    lon = (ll + lu) / 2
                    dt = dt1 + (dt2 - dt1) / 2
                    dt = dt.astimezone(UTC)

                    solar_day = datacube.api.query._convert_to_solar_time(
                        dt, lon
                    ).date()
                    dates.add(solar_day)
            else:
                results = conn.execute(
                    text(
                        """
                  select
                        array_agg(dt.search_val)
                  from odc.dataset_search_datetime dt,
                       odc.dataset ds
                  WHERE ds.product_ref = ANY(:prodids)
                  AND   ds.id = dt.dataset_ref
                  AND   dt.search_key = 'time'
                  """
                    ),
                    {"prodids": prodids},
                )
                for result in results:
                    for dat_ran in result[0]:
                        dates.add(dat_ran.lower)

            if layer.time_resolution.is_subday():
                date_formatter = lambda d: d.isoformat()  # noqa: E731
            else:
                date_formatter = lambda d: d.strftime("%Y-%m-%d")  # noqa: E731

            dates = sorted(dates)
            conn.execute(
                text("""
               UPDATE ows.layer_ranges
               SET dates = :dates
               WHERE layer= :layer_id
            """),
                {
                    "dates": Json(list(map(date_formatter, dates))),
                    "layer_id": layer.name,
                },
            )
            # calculate bounding boxes
            # Get extent polygon from materialised views

            all_bboxes = build_bboxes(layer)
            if not all_bboxes:
                return

            conn.execute(
                text("""
            UPDATE ows.layer_ranges
            SET bboxes = :bbox,
                lat_min = :lat_min,
                lat_max = :lat_max,
                lon_min = :lon_min,
                lon_max = :lon_max
            WHERE layer = :layer_id
            """),
                {
                    "bbox": Json(all_bboxes),
                    "layer_id": layer.name,
                    "lat_min": all_bboxes["EPSG:4326"]["bottom"],
                    "lat_max": all_bboxes["EPSG:4326"]["top"],
                    "lon_min": all_bboxes["EPSG:4326"]["left"],
                    "lon_max": all_bboxes["EPSG:4326"]["right"],
                },
            )

        txn.commit()
        # Only update cache after successful commit.
        if cached is None:
            cache[meta] = []
        cache[meta].append(layer.name)


def build_bboxes(layer: OWSNamedLayer) -> dict[str, dict[str, float]]:
    # User native CRS for base extent, unless no spatial index for it, in which case
    # use 4326 which always has a spatial index.
    base_crs = CRS(layer.native_CRS)
    if base_crs not in layer.dc.index.spatial_indexes():
        base_crs = CRS("epsg:4326")
    base_extent = extent_for_layer(layer, base_crs)
    if base_extent is None:
        click.echo(f"Layer {layer.name} has no extent in CRS {base_crs}.  Skipping.")
        return {}
    result = {}
    for crsid, crs in layer.global_config().crses.items():
        if crs == base_crs:
            result[crsid] = sanitise_bbox(base_extent.boundingbox)
        elif crs in layer.dc.index.spatial_indexes():
            extent = extent_for_layer(layer, crs)
            if extent is None:
                click.echo(f"Layer {layer.name} has no extent in CRS {crs}.  Skipping.")
                return {}
            result[crsid] = sanitise_bbox(extent.boundingbox)
        else:
            extent = base_extent.to_crs(crs)
            result[crsid] = sanitise_bbox(extent.boundingbox)
    return result


def extent_for_layer(layer: OWSNamedLayer, crs: CRS) -> odc.geo.Geometry | None:
    ext = None
    for product in layer.products:
        prod_extent = layer.dc.index.products.spatial_extent(product, crs)
        ext = prod_extent if ext is None else ext | prod_extent
    return ext


def bbox_projections(
    starting_box: odc.geo.Geometry, crses: dict[str, odc.geo.CRS]
) -> dict[str, dict[str, float]]:
    result = {}
    for crsid, crs in crses.items():
        projbbox = starting_box.to_crs(crs).boundingbox
        result[crsid] = sanitise_bbox(projbbox)
    return result


def sanitise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    def sanitise_coordinate(coord: float, fallback: float, upper: bool) -> float:
        if (
            not math.isfinite(coord)
            or (upper and coord > fallback)
            or (not upper and coord < fallback)
        ):
            return fallback
        return coord

    if bbox.crs == CRS("epsg:4326"):
        return {
            "top": sanitise_coordinate(bbox.top, float("90.0"), True),
            "bottom": sanitise_coordinate(bbox.bottom, float("-90.0"), False),
            "left": sanitise_coordinate(bbox.left, float("-180"), False),
            "right": sanitise_coordinate(bbox.right, float("9.999999999e99"), True),
        }
    return {
        "top": sanitise_coordinate(bbox.top, float("9.999999999e99"), True),
        "bottom": sanitise_coordinate(bbox.bottom, float("-9.999999999e99"), False),
        "left": sanitise_coordinate(bbox.left, float("-9.999999999e99"), False),
        "right": sanitise_coordinate(bbox.right, float("9.999999999e99"), True),
    }


def get_ranges(layer: OWSNamedLayer) -> LayerExtent | None:
    with get_sqlconn(layer.dc) as conn:
        results = conn.execute(
            text("""
            SELECT *
            FROM ows.layer_ranges
            WHERE layer=:pname"""),
            {"pname": layer.name},
        )

        for result in results:
            if layer.time_resolution.is_subday():
                dt_parser: Callable[[str], datetime | date] = (  # noqa: E731
                    lambda dts: datetime.fromisoformat(dts)
                )
            else:
                dt_parser = lambda dts: datetime.strptime(dts, "%Y-%m-%d").date()  # noqa: E731
            times = [dt_parser(d) for d in result.dates if d is not None]
            if not times:
                return None
            return LayerExtent(
                lat=CoordRange(min=float(result.lat_min), max=float(result.lat_max)),
                lon=CoordRange(min=float(result.lon_min), max=float(result.lon_max)),
                times=times,
                bboxes=result.bboxes,
            )
        return None



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

from typing import cast

from typing_extensions import override
from xarray import DataArray, Dataset

from datacube_ows.config_utils import CFG_DICT, ConfigException
from datacube_ows.styles.base import StyleDefBase
from datacube_ows.styles.component import LINEAR_COMP_DICT, ComponentStyleDef
from datacube_ows.styles.ramp import ColorRampDef

TYPE_CHECKING = False
if TYPE_CHECKING:
    from datacube_ows.ows_configuration import OWSNamedLayer


class HybridStyleDef(ColorRampDef, ComponentStyleDef):
    """
    Hybrid component/colour ramp style type.

    Returns a linear blend of a component image and colour ramp image
    """

    auto_legend = False

    def __init__(
        self,
        product: "OWSNamedLayer",
        style_cfg: CFG_DICT,
        defer_multi_date: bool = False,
        stand_alone: bool = False,
        user_defined: bool = False,
    ) -> None:
        """
        See StyleBaseDef
        """
        super().__init__(
            product,
            style_cfg,
            defer_multi_date=defer_multi_date,
            stand_alone=stand_alone,
            user_defined=user_defined,
        )
        style_cfg = cast(CFG_DICT, self._raw_cfg)
        self.component_ratio = float(cast(float | str, style_cfg["component_ratio"]))
        if self.component_ratio < 0.0 or self.component_ratio > 1.0:
            raise ConfigException(
                "Component ratio must be a floating point number between 0 and 1"
            )

    @override
    def transform_single_date_data(self, data: Dataset) -> Dataset:
        """
        Apply style to raw data to make an RGBA image xarray (single time slice only)

        :param data: Raw data, all bands.
        :return: RGBA uint8 xarray
        """
        # pylint: disable=too-many-locals
        if self.index_function is not None:
            idx_data = self.index_function(data)
            data["index_function"] = (idx_data.dims, idx_data.data)

        imgdata = Dataset(coords=data)

        d: DataArray = data["index_function"]
        for band, _ in self.rgb_components.items():
            rampdata = DataArray(
                self.color_ramp.get_value(d, band), coords=d.coords, dims=d.dims
            )
            component_band_data: DataArray | None = None
            for c_band, c_intensity in cast(
                LINEAR_COMP_DICT, self.rgb_components[band]
            ).items():
                if callable(c_intensity):
                    imgband_component_data = cast(
                        DataArray, c_intensity(data[c_band], c_band, band)
                    )
                else:
                    imgband_component_data = data[c_band] * cast(DataArray, c_intensity)
                if component_band_data is not None:
                    component_band_data += imgband_component_data
                else:
                    component_band_data = imgband_component_data
                if band != "alpha":
                    component_band_data = self.compress_band(band, component_band_data)
            img_band_data = rampdata * 255.0 * (
                1.0 - self.component_ratio
            ) + self.component_ratio * cast(DataArray, component_band_data)
            imgdata[band] = (d.dims, img_band_data.astype("uint8").data)

        return imgdata


# Register HybridStyleDef as a priority subclass of StyleDefBase
#    (priority means takes precedence over ComponentStyleDef and ColourRampDef)
StyleDefBase.register_subclass(HybridStyleDef, "component_ratio", priority=True)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import datetime
import json
from collections.abc import Iterable
from enum import Enum
from types import UnionType
from typing import cast
from uuid import UUID as UUID_

from datacube import Datacube
from datacube.model import Dataset, Product
from geoalchemy2 import Geometry
from geoalchemy2.functions import ST_AsGeoJSON, ST_Union
from odc.geo.geom import Geometry as ODCGeom
from sqlalchemy import SMALLINT, Column, MetaData, Table, and_, or_, select, text
from sqlalchemy.dialects.postgresql import TSTZRANGE, UUID
from sqlalchemy.engine import Row
from sqlalchemy.sql.elements import ClauseElement
from sqlalchemy.sql.functions import count, func

from datacube_ows.utils import default_to_utc, get_driver_name, get_sqlconn


def get_st_view(meta: MetaData) -> Table:
    return Table(
        "space_time_view",
        meta,
        Column("id", UUID()),
        Column("dataset_type_ref", SMALLINT()),
        Column(
            "spatial_extent", Geometry(from_text="ST_GeomFromGeoJSON", name="geometry")
        ),
        Column("temporal_extent", TSTZRANGE()),
        schema="ows",
    )


_meta = MetaData()
st_view: Table = get_st_view(_meta)


class MVSelectOpts(Enum):
    """
    Enum for mv_search_datasets sel parameter.

    ALL: return all columns, select *, as result set
    IDS: return list of database_ids only.
    DATASETS: return list of ODC dataset objects
    COUNT: return a count of matching datasets
    EXTENT: return full extent of query result as a Geometry
    """

    ALL = 0
    IDS = 1
    COUNT = 2
    EXTENT = 3
    DATASETS = 4

    def sel(self, stv: Table) -> list[ClauseElement]:
        if self == self.ALL:
            return [stv]
        if self == self.IDS or self == self.DATASETS:
            return [stv.c.id]
        if self == self.COUNT:
            return [cast(ClauseElement, count(stv.c.id))]
        if self == self.EXTENT:
            return [ST_AsGeoJSON(ST_Union(text("spatial_extent")))]
        raise AssertionError("Invalid selection option")


selection_return_types: dict[MVSelectOpts, type | UnionType] = {
    MVSelectOpts.ALL: Iterable[Row],
    MVSelectOpts.IDS: Iterable[UUID_],
    MVSelectOpts.DATASETS: Iterable[Dataset],
    MVSelectOpts.COUNT: int,
    MVSelectOpts.EXTENT: ODCGeom | None,
}


SelectOut = Iterable[Row] | Iterable[UUID_] | Iterable[Dataset] | int | ODCGeom | None
DateOrDateTime = datetime.datetime | datetime.date
TimeSearchTerm = (
    tuple[datetime.datetime, datetime.datetime]
    | tuple[datetime.date, datetime.date]
    | DateOrDateTime
)


def mv_search(
    dc: Datacube,
    sel: MVSelectOpts = MVSelectOpts.IDS,
    times: Iterable[TimeSearchTerm] | None = None,
    geom: ODCGeom | None = None,
    products: Iterable[Product] | None = None,
) -> SelectOut:
    """
    Perform a dataset query via the space_time_view

    :param products: An iterable of combinable products to search
    :param index: A datacube index (required)

    :param sel: Selection mode - a MVSelectOpts enum. Defaults to IDS.
    :param times: A list of pairs of datetimes (with time zone)
    :param geom: A odc.geo.geom.Geometry object

    :return: See MVSelectOpts doc
    """
    stv = st_view
    if products is None:
        raise Exception("Must filter by product/layer")
    prod_ids = [p.id for p in products]

    s = select(*sel.sel(stv)).where(stv.c.dataset_type_ref.in_(prod_ids))  # type: ignore[call-overload]
    if times is not None:
        or_clauses = []
        for t in times:
            if isinstance(t, datetime.datetime):
                st = default_to_utc(t)
                tmax = st + datetime.timedelta(seconds=1)
                or_clauses.append(
                    and_(
                        func.lower(stv.c.temporal_extent) >= t,
                        func.lower(stv.c.temporal_extent) < tmax,
                    )
                )
            elif isinstance(t, datetime.date):
                st = datetime.datetime(t.year, t.month, t.day, tzinfo=datetime.UTC)
                tmax = st + datetime.timedelta(days=1)
                or_clauses.append(
                    and_(
                        func.lower(stv.c.temporal_extent) >= st,
                        func.lower(stv.c.temporal_extent) < tmax,
                    )
                )
            else:
                if get_driver_name(dc.index) == "psycopg":
                    from psycopg.types.range import TimestamptzRange

                    TZRange = TimestamptzRange
                else:
                    from psycopg2.extras import DateTimeTZRange

                    TZRange = DateTimeTZRange
                or_clauses.append(stv.c.temporal_extent.op("&&")(TZRange(*t)))
        s = s.where(or_(*or_clauses))
    orig_crs = None
    if geom is not None:
        orig_crs = geom.crs
        if str(geom.crs) != "EPSG:4326":
            geom = geom.to_crs("EPSG:4326")
        geom_js = json.dumps(geom.json)
        s = s.where(stv.c.spatial_extent.intersects(geom_js))
    # print(s) # Print SQL Statement
    if sel == MVSelectOpts.ALL:
        with get_sqlconn(dc) as conn:
            return conn.execute(s)
    if sel == MVSelectOpts.IDS:
        with get_sqlconn(dc) as conn:
            return [r[0] for r in conn.execute(s)]
    if sel in (MVSelectOpts.COUNT, MVSelectOpts.EXTENT):
        with get_sqlconn(dc) as conn:
            for r in conn.execute(s):
                if sel == MVSelectOpts.COUNT:
                    return cast(int, r[0])
                # MVSelectOpts.EXTENT
                geojson = r[0]
                if geojson is None:
                    return None
                uniongeom = ODCGeom(json.loads(geojson), crs="EPSG:4326")
                if geom:
                    intersect = uniongeom.intersection(geom)
                    if intersect.wkt == "POLYGON EMPTY":
                        return None
                    if orig_crs and orig_crs != "EPSG:4326":
                        intersect = intersect.to_crs(orig_crs)
                else:
                    intersect = uniongeom
                return intersect
    elif sel == MVSelectOpts.DATASETS:
        with get_sqlconn(dc) as conn:
            ids = [r[0] for r in conn.execute(s)]
        return dc.index.datasets.bulk_get(ids)
    else:
        raise Exception("Invalid Selection Option")
    raise Exception("Unreachable code reached")



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import datetime
from collections.abc import Iterable
from threading import Lock
from typing import Any, Literal
from uuid import UUID

import click
from antimeridian import fix_shape
from datacube import Datacube
from datacube.drivers.common_psql import as_role
from datacube.model import Dataset, Product, Range
from odc.geo import CRS, Geometry
from sqlalchemy import text
from typing_extensions import override

from datacube_ows.index.api import (
    AbortRun,
    InsufficientDbPrivileges,
    LayerExtent,
    LayerSignature,
    OWSAbstractIndex,
    OWSAbstractIndexDriver,
    TimeSearchTerm,
    check_perms,
)
from datacube_ows.index.sql import run_sql
from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.utils import get_driver_name, get_sqlconn

from ...utils import default_to_utc
from .product_ranges import create_range_entry as create_range_entry_impl
from .product_ranges import get_ranges as get_ranges_impl


class OWSPostgisIndex(OWSAbstractIndex):
    name: str = "postgis"

    # method to check database access (for ping op)
    @override
    def check_db_access(self, dc: Datacube) -> bool:
        db_ok = False
        try:
            with get_sqlconn(dc) as conn:
                results = conn.execute(
                    text("""
                    SELECT *
                    FROM ows.layer_ranges
                    LIMIT 1""")
                )
                for _ in results:
                    db_ok = True
        except Exception:
            pass
        return db_ok

    @override
    def _check_perms(self, dc: Datacube, group: Literal["manage", "admin"]) -> None:
        if get_driver_name(dc.index) == "psycopg":
            from psycopg.errors import ProgrammingError
        else:
            from psycopg2.errors import ProgrammingError
        try:
            with get_sqlconn(dc) as conn, as_role(conn, f"odc_{group}"):
                pass
        except ProgrammingError:
            raise InsufficientDbPrivileges(
                f"db user {dc.index.environment.db_username} does not have odc_{group} privileges"
            ) from None

    # method to delete obsolete schemas etc.
    @override
    @check_perms("admin")
    def cleanup_schema(self, dc: Datacube) -> None:
        # No obsolete schema for postgis databases to clean up.
        pass

    # Schema creation method
    @override
    @check_perms("admin")
    def create_schema(self, dc: Datacube) -> None:
        click.echo("Creating schema...")
        if not self._run_sql(dc, "ows_schema/bootstrap"):
            raise AbortRun(
                "Could not bootstrap schema: "
                "try using an ODC environment that connects as a database superuser."
            )
        click.echo("Creating/updating tables...")
        self._run_sql(dc, "ows_schema/create")
        click.echo("Granting tables permissions to odc roles...")
        self._run_sql(dc, "ows_schema/grants/")

    # Spatiotemporal index update method (e.g. refresh materialised views)
    @override
    @check_perms("manage")
    def update_geotemporal_index(self, dc: Datacube) -> None:
        # Native ODC geotemporal index used in postgis driver.
        pass

    @override
    @check_perms("manage")
    def create_range_entry(
        self, layer: OWSNamedLayer, cache: dict[LayerSignature, list[str]]
    ) -> None:
        create_range_entry_impl(layer, cache)

    @override
    def get_ranges(self, layer: OWSNamedLayer) -> LayerExtent | None:
        return get_ranges_impl(layer)

    def _query(
        self,
        layer: OWSNamedLayer,
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
    ) -> dict[str, Any]:
        query: dict[str, Any] = {}
        if geom:
            if geom.crs and geom.crs in layer.dc.index.spatial_indexes():
                query["geopolygon"] = geom
            else:
                # Default to 4326 and take a long hard look at yourself.
                prepared_geom = self._prep_geom(layer, geom)
                assert prepared_geom is not None
                geopoly = prepared_geom.to_crs("epsg:4326")
                geopoly = Geometry(fix_shape(geopoly.geom), crs="epsg:4326")
                query["geopolygon"] = geopoly
        if products is not None:
            query["product"] = [p.name for p in products]
        if times is not None:

            def normalise_to_dtr(
                unnorm: datetime.datetime | datetime.date,
            ) -> tuple[datetime.datetime, datetime.datetime]:
                if isinstance(unnorm, datetime.datetime):
                    st: datetime.datetime = default_to_utc(unnorm)
                    tmax = st + datetime.timedelta(seconds=1)
                elif isinstance(t, datetime.date):
                    st = datetime.datetime(
                        unnorm.year,
                        unnorm.month,
                        unnorm.day,
                        tzinfo=datetime.UTC,
                    )
                    tmax = st + datetime.timedelta(days=1)
                else:
                    raise ValueError("Not a datetime object")
                return st, tmax

            time_args = []
            for t in times:
                if isinstance(t, datetime.date | datetime.datetime):
                    start, tmax = normalise_to_dtr(t)
                    time_args.append(Range(start, tmax))
                else:
                    st, et = t
                    st, _ = normalise_to_dtr(st)
                    et, _ = normalise_to_dtr(et)
                    time_args.append(Range(st, et))
            query["time"] = time_args[0]
        return query

    @override
    def ds_search(
        self,
        layer: OWSNamedLayer,
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
    ) -> Iterable[Dataset]:
        return layer.dc.index.datasets.search(
            **self._query(layer, times, geom, products)
        )

    @override
    def dsid_search(
        self,
        layer: OWSNamedLayer,
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
    ) -> Iterable[UUID]:
        for ds in layer.dc.index.datasets.search_returning(
            field_names=["id"], **self._query(layer, times, geom, products)
        ):
            yield ds.id  # type: ignore[attr-defined]

    @override
    def count(
        self,
        layer: OWSNamedLayer,
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
    ) -> int:
        return layer.dc.index.datasets.count(
            **self._query(layer, times, geom, products)
        )

    @override
    def extent(
        self,
        layer: OWSNamedLayer,
        times: Iterable[TimeSearchTerm] | None = None,
        geom: Geometry | None = None,
        products: Iterable[Product] | None = None,
        crs: CRS | None = None,
    ) -> Geometry | None:
        if crs is None:
            crs = CRS("epsg:4326")
        return layer.dc.index.datasets.spatial_extent(
            self.dsid_search(layer, times=times, geom=geom, products=products), crs=crs
        )

    def _run_sql(self, dc: Datacube, path: str, **params: str) -> bool:
        return run_sql(dc, path, **params)


pgisdriverlock = Lock()


class OWSPostgisIndexDriver(OWSAbstractIndexDriver):
    _driver = None

    @classmethod
    @override
    def ows_index_class(cls) -> type[OWSAbstractIndex]:
        return OWSPostgisIndex

    @classmethod
    @override
    def ows_index(cls) -> OWSAbstractIndex:
        with pgisdriverlock:
            if cls._driver is None:
                cls._driver = OWSPostgisIndex()
        return cls._driver


def ows_index_driver_init() -> OWSPostgisIndexDriver:
    return OWSPostgisIndexDriver()



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
import sys
import traceback
from logging import Logger

from flask import render_template, request
from sqlalchemy.exc import OperationalError

from datacube_ows import __version__
from datacube_ows.http_utils import (
    capture_headers,
    get_service_base_url,
    lower_get_args,
    resp_headers,
)
from datacube_ows.index.api import ows_index
from datacube_ows.legend_generator import create_legend_for_style
from datacube_ows.ogc_exceptions import OGCException, WMSException
from datacube_ows.ows_configuration import get_config
from datacube_ows.protocol_versions import SupportedSvc, supported_versions
from datacube_ows.wcs1 import WCS_REQUESTS
from datacube_ows.wms import WMS_REQUESTS

_LOG: Logger = logging.getLogger(__name__)

# Protocol/Version lookup table
OWS_SUPPORTED: dict[str, SupportedSvc] = supported_versions()


# Flask Routes
# (Note that actual route declarations take place in startup_utils.py)
def ogc_impl():
    # pylint: disable=too-many-branches
    nocase_args = capture_headers(request, lower_get_args())
    service = nocase_args.get("service", "").upper()

    if service:
        return ogc_svc_impl(service.lower())

    # create dummy env if not exists
    try:
        # service argument is only required (in fact only defined) by OGC for
        # GetCapabilities requests.  As long as we are persisting with a single
        # routing end point for all services, we must derive the service from the request
        # parameter.
        # This is a quick hack to fix #64.  Service and operation routing could be
        # handled more elegantly.
        op = nocase_args.get("request", "").upper()
        if op in WMS_REQUESTS:
            return ogc_svc_impl("wms")
        if op in WCS_REQUESTS:
            return ogc_svc_impl("wcs")
        if op:
            # Should we return a WMS or WCS exception if there is no service specified?
            # Defaulting to WMS because that's what we already have.
            raise WMSException(
                "Invalid service and/or request",
                locator="Service and request parameters",
            )
        cfg = get_config()  # pylint: disable=redefined-outer-name
        url = nocase_args.get("Host", nocase_args["url_root"])
        base_url = get_service_base_url(cfg.allowed_urls, url)
        return (
            render_template(
                "index.html",
                cfg=cfg,
                supported=OWS_SUPPORTED,
                base_url=base_url,
                version=__version__,
            ),
            200,
            resp_headers({"Content-Type": "text/html"}),
        )
    except OGCException as e:
        _LOG.error("Handled Error: %s", repr(e.errors))
        return e.exception_response()
    except Exception as e:  # pylint: disable=broad-except
        _LOG.exception(e)
        tb = sys.exc_info()[2]
        ogc_e = WMSException(f"Unexpected server error: {e!s}", http_response=500)
        return ogc_e.exception_response(traceback=traceback.extract_tb(tb))


def ogc_svc_impl(svc):
    svc_support = OWS_SUPPORTED.get(svc)
    nocase_args = capture_headers(request, lower_get_args())
    service = nocase_args.get("service", svc).upper()

    # Is service activated in config?
    try:
        if not svc_support:
            raise WMSException(
                f"Invalid service: {svc}",
                valid_keys=[
                    service.service
                    for service in OWS_SUPPORTED.values()
                    if service.activated()
                ],
                code=WMSException.OPERATION_NOT_SUPPORTED,
                locator="service parameter",
            )
        if not svc_support.activated():
            raise svc_support.default_exception_class(
                "Invalid service and/or request",
                locator="Service and request parameters",
            )

        # Does service match path (if supplied)
        if service != svc_support.service_upper:
            raise svc_support.default_exception_class(
                "Invalid service", locator="Service parameter"
            )

        version = nocase_args.get("version")
        version_support = svc_support.negotiated_version(version)
    except OGCException as e:
        return e.exception_response()

    try:
        return version_support.router(nocase_args)
    except OGCException as e:
        return e.exception_response()
    except OperationalError as e:
        # Expected types of failures in this branch. Log the error to console and give
        # the user a generic 500 response.
        _LOG.error(e)
        return version_support.exception_class(
            "Internal server error.", http_response=500
        ).exception_response()
    except Exception as e:  # pylint: disable=broad-except
        _LOG.exception(e)
        tb = sys.exc_info()[2]
        ogc_e = version_support.exception_class(
            f"Unexpected server error: {e!s}", http_response=500
        )
        return ogc_e.exception_response(traceback=traceback.extract_tb(tb))


def ogc_wms_impl():
    return ogc_svc_impl("wms")


def ogc_wmts_impl():
    return ogc_svc_impl("wmts")


def ogc_wcs_impl():
    return ogc_svc_impl("wcs")


def ping() -> tuple[str, int, dict[str, str]]:
    dbs_ok = {
        name: ows_index(dc).check_db_access(dc)
        for name, dc in get_config().all_dcs.items()
    }

    if all(dbs_ok.values()):
        return (
            render_template("ping.html", status="Up", statuses=dbs_ok),
            200,
            resp_headers({"Content-Type": "text/html"}),
        )
    if any(dbs_ok.values()):
        return (
            render_template("ping.html", status="Partially Up", statuses=dbs_ok),
            503,
            resp_headers({"Content-Type": "text/html"}),
        )
    return (
        render_template("ping.html", status="Down", statuses=dbs_ok),
        503,
        resp_headers({"Content-Type": "text/html"}),
    )


def legend(layer, style, dates=None):
    # pylint: disable=redefined-outer-name
    cfg = get_config()
    product = cfg.layer_index.get(layer)
    if not product:
        return "Unknown Layer", 404, resp_headers({"Content-Type": "text/plain"})
    ndates = int(lower_get_args().get("ndates", 0)) if dates is None else len(dates)
    try:
        img = create_legend_for_style(product, style, ndates)
    except WMSException as e:
        return str(e), e.http_response, resp_headers({"Content-Type": "text/plain"})

    if not img:
        return "Unknown Style", 404, resp_headers({"Content-Type": "text/plain"})
    return img



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0


# pylint: skip-file
import os
import sys

# This is the directory of the source code that the web app will run from
sys.path.append("/src")

# The location of the datacube config file.
if os.path.isfile("/src/odc/.datacube.conf.local"):
    os.environ.setdefault("ODC_CONFIG_PATH", "/src/odc/.datacube.conf.local")

from datacube_ows import __version__, create_app

application = create_app()


def main() -> None:
    if "--version" in sys.argv:
        print("Open Data Cube Open Web Services (datacube-ows) version", __version__)
        exit(0)
    application.run()


if __name__ == "__main__":
    main()



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import math
from collections.abc import Iterable, Mapping
from typing import Any, cast

import affine
import numpy as np
from odc.geo.crs import CRS
from odc.geo.geobox import GeoBox
from odc.geo.geom import polygon

from datacube_ows.config_utils import CFG_DICT, RAW_CFG, ConfigException, OWSConfigEntry
from datacube_ows.http_utils import cache_control_headers
from datacube_ows.ogc_utils import create_geobox

TYPE_CHECKING = False
if TYPE_CHECKING:
    import datacube_ows.ows_configuration


def parse_cache_age(cfg: dict, entry, section, default: int = 0) -> int:
    try:
        val = int(cfg.get(entry, default))
    except ValueError:
        raise ConfigException(
            f"{entry} in {section} section must be an integer: {cfg[entry]}"
        ) from None
    if val < 0:
        raise ConfigException(
            f"{entry} in {section} section cannot be negative: {cfg[entry]}"
        )
    return val


# pyre-ignore[13]
class RequestScale:
    standard_scale: "RequestScale"

    def __init__(
        self,
        native_crs: CRS,
        native_resolution: tuple[float | int, float | int],
        geobox: GeoBox,
        n_dates: int,
        request_bands: Iterable[Mapping[str, Any]] | None = None,
        total_band_size: int | None = None,
    ) -> None:
        self.resolution = self._metre_resolution(native_crs, native_resolution)
        self.crs = native_crs
        self.geobox = self._standardise_geobox(geobox)
        self.pixel_size = (geobox.width, geobox.height)
        self.n_dates = n_dates
        self.bands = request_bands
        assert (request_bands is not None) ^ (total_band_size is not None)
        if total_band_size is not None:
            self.total_band_size: int = total_band_size
        else:
            self.total_band_size = sum(
                np.dtype(band["dtype"]).itemsize
                for band in cast(Iterable[Mapping[str, Any]], request_bands)
            )

    def _standardise_geobox(self, geobox: GeoBox) -> GeoBox:
        if geobox.crs == "EPSG:3857":
            return geobox
        bbox = geobox.extent.to_crs("EPSG:3857").boundingbox
        return create_geobox(
            CRS("EPSG:3857"),
            bbox.left,
            bbox.bottom,
            bbox.right,
            bbox.top,
            width=geobox.width,
            height=geobox.height,
        )

    def _metre_resolution(
        self, crs: CRS, resolution: tuple[float | int, float | int]
    ) -> tuple[float, float]:
        # Convert native resolution to metres for ready comparison.
        if crs.units == ("metre", "metre"):
            return cast(tuple[float, float], tuple(abs(r) for r in resolution))
        resolution_rectangle = polygon(
            [(0, 0), (0, resolution[1]), resolution, (0, resolution[0]), (0, 0)],
            crs=crs,
        )
        proj_bbox = resolution_rectangle.to_crs("EPSG:3857").boundingbox
        return (
            abs(proj_bbox.right - proj_bbox.left),
            abs(proj_bbox.top - proj_bbox.bottom),
        )

    def pixel_span(self) -> tuple[float, float]:
        bbox = self.geobox.extent.boundingbox
        return (
            (bbox.right - bbox.left) / self.geobox.width,
            (bbox.bottom - bbox.top) / self.geobox.height,
        )

    @property
    def scale_denominator(self) -> float:
        xy_denoms = [abs(ps / 0.00028) for ps in self.pixel_span()]
        return sum(xy_denoms) / 2.0

    @property
    def base_zoom_level(self) -> float:
        return math.log2(559082264.0287178 / self.scale_denominator)

    @property
    def load_adjusted_zoom_level(self) -> float:
        return self.base_zoom_level - self.zoom_lvl_offset

    def res_xy(self) -> int | float:
        return self.resolution[0] * self.resolution[1]

    def __truediv__(self, other: "RequestScale") -> float:
        ratio = 1.0
        ratio = ratio * self.n_dates / other.n_dates
        for i in range(2):
            ratio = ratio * self.pixel_size[i] / other.pixel_size[i]
        ratio = ratio * self.total_band_size / other.total_band_size
        return ratio * other.res_xy() / self.res_xy()

    @property
    def load_factor(self) -> float:
        return self / self.standard_scale

    @property
    def zoom_lvl_offset(self) -> float:
        return math.log(self.load_factor, 4)


RequestScale.standard_scale = RequestScale(
    CRS("EPSG:3857"),
    (25.0, 25.0),
    GeoBox((256, 256), affine=affine.identity, crs="EPSG:3857"),
    1,
    total_band_size=(3 * 2),
)


class CacheControlRules(OWSConfigEntry):
    def __init__(self, cfg: RAW_CFG, context: str, max_datasets: int) -> None:
        """
        Class constructor. stores and validates rule dictionaries

        :param cfg: None, or a list of progressive CCR configurations.
        :param context: Context label (e.g. owning layer name), used for reporting validation errors
        :param max_datasets: Over-arching maximum dataset limit in context.
        """
        super().__init__(cfg)
        self.rules = cast(list[CFG_DICT] | None, self._raw_cfg)
        self.use_caching: bool = self.rules is not None
        self.max_datasets = max_datasets
        if not self.use_caching:
            return

        # Validate rules
        min_so_far: int = 0
        max_max_age_so_far: int = 0
        for rule in cast(list[CFG_DICT], self.rules):
            if "min_datasets" not in rule:
                raise ConfigException(
                    f"Dataset cache rule does not contain a 'min_datasets' element in {context}"
                )
            if "max_age" not in rule:
                raise ConfigException(
                    f"Dataset cache rule does not contain a 'max_age' element in {context}"
                )
            if not isinstance(rule["min_datasets"], int):
                raise ConfigException(
                    f"Dataset cache rule has non-integer 'min_datasets' element in {context}"
                )
            min_datasets = rule["min_datasets"]
            if not isinstance(rule["max_age"], int):
                raise ConfigException(
                    f"Dataset cache rule has non-integer 'max_age' element in {context}"
                )
            max_age = rule["max_age"]
            if min_datasets <= 0:
                raise ConfigException(
                    f"Invalid dataset cache rule in {context}: min_datasets must be greater than zero."
                )
            if min_datasets <= min_so_far:
                raise ConfigException(
                    f"Dataset cache rules must be sorted by ascending min_datasets values.  In layer {context}."
                )
            if max_datasets > 0 and min_datasets > max_datasets:
                raise ConfigException(
                    f"Dataset cache rule min_datasets value exceeds the max_datasets limit in layer {context}."
                )
            min_so_far = min_datasets
            if max_age <= 0:
                raise ConfigException(
                    f"Dataset cache rule max_age value must be greater than zero in layer {context}."
                )
            if max_age <= max_max_age_so_far:
                raise ConfigException(
                    f"max_age values in dataset cache rules must increase monotonically in layer {context}."
                )
            max_max_age_so_far = max_age

    def cache_headers(self, n_datasets: int) -> Mapping[str, str]:
        """
        Generate Cache Control headers for a request accessing a number of datasets.

        :param n_datasets: number of datasets
        :return: Dictionary containing the appropriate HTTP cache control response headers.
        """
        if not self.use_caching:
            return {}
        assert n_datasets >= 0
        # If there are no datasets, don't cache. But if the max_datasets isn't 0, check
        # we're not exceeding it, and in that case, don't cache either.
        if n_datasets == 0 or (
            self.max_datasets > 0 and n_datasets > self.max_datasets
        ):
            return cache_control_headers(0)
        rule = None
        for r in cast(list[CFG_DICT], self.rules):
            if n_datasets < cast(int, r["min_datasets"]):
                break
            rule = r
        if rule:
            return cache_control_headers(cast(int, rule["max_age"]))
        return cache_control_headers(0)


class ResourceLimited(Exception):
    def __init__(self, reasons: list[str], wcs_hard: bool = False) -> None:
        self.reasons = reasons
        self.wcs_hard = wcs_hard
        super().__init__(f"Resource limit(s) exceeded: {','.join(reasons)}")


class OWSResourceManagementRules(OWSConfigEntry):
    # pylint: disable=attribute-defined-outside-init
    def __init__(
        self,
        global_cfg: "datacube_ows.ows_configuration.OWSConfig",
        cfg: CFG_DICT,
        context: str,
    ) -> None:
        """
        Class constructor.

        :param cfg: A resource limit configuration dictionary.
        :param context: The context (e.g. layer name) for reporting validation errors.
        """
        super().__init__(cfg)
        cfg = cast(CFG_DICT, self._raw_cfg)
        wms_cfg = cast(CFG_DICT, cfg.get("wms", {}))
        wcs_cfg = cast(CFG_DICT, cfg.get("wcs", {}))
        self.zoom_fill = cast(
            list[int], wms_cfg.get("zoomed_out_fill_colour", [150, 180, 200, 160])
        )
        if len(self.zoom_fill) == 3:
            self.zoom_fill += [255]
        if len(self.zoom_fill) != 4:
            raise ConfigException(
                f"zoomed_out_fill_colour must have 3 or 4 elements in {context}"
            )
        self.min_zoom = cast(float | None, wms_cfg.get("min_zoom_factor"))
        self.min_zoom_lvl = cast(int | float | None, wms_cfg.get("min_zoom_level"))
        self.max_datasets_wms = cast(int, wms_cfg.get("max_datasets", 0))
        self.max_datasets_wcs = cast(int, wcs_cfg.get("max_datasets", 0))
        self.max_image_size_wcs = cast(int, wcs_cfg.get("max_image_size", 0))
        self.wms_cache_rules = CacheControlRules(
            wms_cfg.get("dataset_cache_rules"), context, self.max_datasets_wms
        )
        self.wcs_cache_rules = CacheControlRules(
            wcs_cfg.get("dataset_cache_rules"), context, self.max_datasets_wcs
        )
        self.wcs_desc_cache_rule = parse_cache_age(
            cfg,
            "describe_cache_max_age",
            f"resource_limits for {context}",
            default=global_cfg.wcs_default_descov_age,
        )

    def check_wms(
        self, n_datasets: int, zoom_factor: float, request_scale: RequestScale
    ) -> None:
        """
        Check whether a WMS requests exceeds the configured resource limits.

        :param n_datasets: The number of datasets for the query
        :param zoom_factor: The zoom factor of the query
        :param request_scale: Model of the resource-intensiveness of the query
        :raises: ResourceLimited if any limits are exceeded.
        """
        limits_exceeded: list[str] = []
        if self.max_datasets_wms > 0 and n_datasets > self.max_datasets_wms:
            limits_exceeded.append("too many datasets")
        if self.min_zoom is not None and zoom_factor < self.min_zoom:
            limits_exceeded.append("zoomed out too far")
        if self.min_zoom_lvl is not None:
            fuzz_factor = 0.01
            if request_scale.load_adjusted_zoom_level < self.min_zoom_lvl - fuzz_factor:
                limits_exceeded.append("too much projected resource requirements")
        if limits_exceeded:
            raise ResourceLimited(limits_exceeded)

    def check_wcs(
        self, n_datasets: int, height: int, width: int, pixel_size: int, n_dates: int
    ) -> None:
        """
        Check whether a WCS requests exceeds the configured resource limits.

        :param n_datasets: The number of datasets for the query
        :raises: ResourceLimited if any limits are exceeded.
        """
        limits_exceeded: list[str] = []
        hard = False
        if self.max_datasets_wcs > 0 and n_datasets > self.max_datasets_wcs:
            limits_exceeded.append(
                f"too many datasets ({n_datasets}: maximum={self.max_datasets_wcs}"
            )
        pixel_count = height * width
        if (
            self.max_image_size_wcs > 0
            and n_dates * pixel_count * pixel_size > self.max_image_size_wcs
        ):
            limits_exceeded.append(
                "too much data for a single request - try selecting fewer pixels or bands"
            )
            hard = True
        if limits_exceeded:
            raise ResourceLimited(limits_exceeded, wcs_hard=hard)



# This file is part of datacube-ows, part of the Open Data Cube project.
# See https://opendatacube.org for more information.
#
# Copyright (c) 2017-2024 OWS Contributors
# SPDX-License-Identifier: Apache-2.0

import logging
import math
from collections.abc import Callable
from datetime import UTC, date, datetime
from typing import cast

import datacube
import odc.geo
from odc.geo.crs import CRS
from odc.geo.geom import Geometry
from sqlalchemy import text

from datacube_ows.index.api import CoordRange, LayerExtent, LayerSignature
from datacube_ows.index.postgres.mv_index import MVSelectOpts, mv_search
from datacube_ows.ows_configuration import OWSNamedLayer
from datacube_ows.utils import get_driver_name, get_sqlconn

_LOG: logging.Logger = logging.getLogger(__name__)


def jsonise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    if isinstance(bbox, dict):
        return bbox
    return {
        "top": bbox.top,
        "bottom": bbox.bottom,
        "left": bbox.left,
        "right": bbox.right,
    }


def create_range_entry(
    layer: OWSNamedLayer, cache: dict[LayerSignature, list[str]]
) -> None:
    # Take layer signature
    meta = LayerSignature(
        time_res=layer.time_resolution.value,
        products=tuple(layer.product_names),
        env=layer.local_env._name,
        datasets=layer.dc.index.datasets.count(product=layer.product_names),
    )

    with get_sqlconn(layer.dc) as conn:
        txn = conn.begin()
        cached = cache.get(meta)
        if get_driver_name(layer.dc.index) == "psycopg":
            from psycopg.types.json import Json
        else:
            from psycopg2.extras import Json
        # insert initial empty row if one does not already exist, so we can safely update it later.
        conn.execute(
            text("""
                 INSERT INTO ows.layer_ranges
                 (layer, lat_min, lat_max, lon_min, lon_max, dates, bboxes, meta, last_updated)
                 VALUES (:p_layer, 0, 0, 0, 0, :empty, :empty, :meta, :now) ON CONFLICT (layer) DO NOTHING
                 """),
            {
                "p_layer": layer.name,
                "empty": Json(""),
                "meta": Json(meta.as_json()),
                "now": datetime.now(tz=UTC),
            },
        )
        if cached is not None:
            template = cached[0]
            # A layer has same signature - reuse range"
            conn.execute(
                text("""
            UPDATE ows.layer_ranges lr1
            SET lat_min = lr2.lat_min,
                lat_max = lr2.lat_max,
                lon_min = lr2.lon_min,
                lon_max = lr2.lon_max,
                dates = lr2.dates,
                bboxes = lr2.bboxes,
                meta = lr2.meta,
                last_updated = lr2.last_updated
            FROM ows.layer_ranges lr2
            WHERE lr1.layer = :layer_id
            AND   lr2.layer = :template_id"""),
                {"layer_id": layer.name, "template_id": template},
            )
        else:
            prodids = [p.id for p in layer.products]
            # Update min/max lat/longs
            conn.execute(
                text(
                    """
              UPDATE ows.layer_ranges lr
              SET lat_min = st_ymin(subq.bbox),
                  lat_max = st_ymax(subq.bbox),
                  lon_min = st_xmin(subq.bbox),
                  lon_max = st_xmax(subq.bbox)
              FROM (
                SELECT st_extent(stv.spatial_extent) as bbox
                FROM ows.space_time_view stv
                WHERE stv.dataset_type_ref = ANY(:prodids)
              ) as subq
              WHERE lr.layer = :layer_id
              """
                ),
                {"layer_id": layer.name, "prodids": prodids},
            )

            # Set default timezone
            conn.execute(text("""set timezone to 'Etc/UTC'"""))

            # Loop over dates
            dates = set()  # Should get to here!
            if layer.time_resolution.is_solar():
                results = conn.execute(
                    text(
                        """
                  select
                        lower(temporal_extent), upper(temporal_extent),
                        ST_X(ST_Centroid(spatial_extent))
                  from ows.space_time_view
                  WHERE dataset_type_ref = ANY(:prodids)
                  """
                    ),
                    {"prodids": prodids},
                )
                for result in results:
                    dt1, dt2, lon = result
                    dt = dt1 + (dt2 - dt1) / 2
                    dt = dt.astimezone(UTC)

                    solar_day = datacube.api.query._convert_to_solar_time(
                        dt, lon
                    ).date()
                    dates.add(solar_day)
            else:
                results = conn.execute(
                    text(
                        """
                  select
                        array_agg(temporal_extent)
                  from ows.space_time_view
                  WHERE dataset_type_ref = ANY(:prodids)
                  """
                    ),
                    {"prodids": prodids},
                )
                for result in results:
                    for dat_ran in result[0]:
                        dates.add(dat_ran.lower)

            if layer.time_resolution.is_subday():
                date_formatter = lambda d: d.isoformat()  # noqa: E731
            else:
                date_formatter = lambda d: d.strftime("%Y-%m-%d")  # noqa: E731

            dates = sorted(dates)
            conn.execute(
                text("""
               UPDATE ows.layer_ranges
               SET dates = :dates
               WHERE layer= :layer_id
            """),
                {
                    "dates": Json(list(map(date_formatter, dates))),
                    "layer_id": layer.name,
                },
            )
            # calculate bounding boxes
            # Get extent polygon from materialised views

            extent_4386 = cast(
                Geometry,
                mv_search(layer.dc, MVSelectOpts.EXTENT, products=layer.products),
            )

            all_bboxes = bbox_projections(extent_4386, layer.global_cfg.crses)

            conn.execute(
                text("""
            UPDATE ows.layer_ranges
            SET bboxes = :bbox
            WHERE layer = :layer_id
            """),
                {"bbox": Json(all_bboxes), "layer_id": layer.name},
            )

        txn.commit()
        # Only update cache after successful commit.
        if cached is None:
            cache[meta] = []
        cache[meta].append(layer.name)


def bbox_projections(
    starting_box: odc.geo.Geometry, crses: dict[str, CRS]
) -> dict[str, dict[str, float]]:
    result = {}
    for crsid, crs in crses.items():
        if crs.valid_region is not None:
            clipped_crs_region = starting_box & crs.valid_region
            if clipped_crs_region.wkt == "POLYGON EMPTY":
                continue
            clipped_crs_bbox = clipped_crs_region.to_crs(crs).boundingbox
        else:
            clipped_crs_bbox = None
        if clipped_crs_bbox is not None:
            result[crsid] = jsonise_bbox(clipped_crs_bbox)
        else:
            projbbox = starting_box.to_crs(crs).boundingbox
            result[crsid] = sanitise_bbox(projbbox)
    return result


def sanitise_bbox(bbox: odc.geo.geom.BoundingBox) -> dict[str, float]:
    def sanitise_coordinate(coord: float, fallback: float) -> float:
        return coord if math.isfinite(coord) else fallback

    return {
        "top": sanitise_coordinate(bbox.top, float("9.999999999e99")),
        "bottom": sanitise_coordinate(bbox.bottom, float("-9.999999999e99")),
        "left": sanitise_coordinate(bbox.left, float("-9.999999999e99")),
        "right": sanitise_coordinate(bbox.right, float("9.999999999e99")),
    }


def get_ranges(layer: OWSNamedLayer) -> LayerExtent | None:
    with get_sqlconn(layer.dc) as conn:
        results = conn.execute(
            text("""
            SELECT *
            FROM ows.layer_ranges
            WHERE layer=:pname"""),
            {"pname": layer.name},
        )

        for result in results:
            if layer.time_resolution.is_subday():
                dt_parser: Callable[[str], datetime | date] = (  # noqa: E731
                    lambda dts: datetime.fromisoformat(dts)
                )
            else:
                dt_parser = lambda dts: datetime.strptime(dts, "%Y-%m-%d").date()  # noqa: E731
            times = [dt_parser(d) for d in result.dates if d is not None]
            if not times:
                return None
            return LayerExtent(
                lat=CoordRange(min=float(result.lat_min), max=float(result.lat_max)),
                lon=CoordRange(min=float(result.lon_min), max=float(result.lon_max)),
                times=times,
                bboxes=result.bboxes,
            )
        return None
